let size1, size2, size3 = 250800000, 106000000, 400000;;
(*let size1, size2, size3 = 705000, 400000, 40000;;*)

let rxp_space = Str.regexp " ";;
let ssplit = Str.split rxp_space;;

let f2i, i2f = Hashtbl.create 1000, Hashtbl.create 1000;;
let tags = Hashtbl.create 1000;;
let hidden = Hashtbl.create 1000;;

let addfact s n =
  if s = "!" then Hashtbl.add tags n ()
  else if String.length s > 6 && String.sub s 0 7 = "hidden_" then Hashtbl.add hidden n ()
  else (Hashtbl.add f2i s n; Hashtbl.add i2f n s);;

let read_facts () =
  let inc = open_in "facts.lst" in
  try while true do
    match ssplit (input_line inc) with
    | [n; i] ->  addfact n (int_of_string i)
    | _ -> assert false
  done with End_of_file -> close_in inc;;

let tyno, tmno, thno = ref 0, ref 0, ref 0;;
let ios s = abs (int_of_string s);;

let process thth thtm tmtm thty tmty tyty = function
  ('R', [t]) -> incr thno; thtm (ios t)
| ('B', [t]) -> incr thno; thtm (ios t)
| ('T', [p; q]) -> incr thno; thth (ios p); thth (ios q)
| ('C', [p; q]) -> incr thno; thth (ios p); thth (ios q)
| ('L', [t; p]) -> incr thno; thth (ios p); thtm (ios t)
| ('H', [t]) -> incr thno; thtm (ios t)
| ('E', [p; q]) -> incr thno; thth (ios p); thth (ios q)
| ('D', [p; q]) -> incr thno; thth (ios p); thth (ios q)
| ('Q', ((h :: t) as l)) -> incr thno; thth (ios (List.hd (List.rev l)));
    List.iter thty (List.map ios (List.tl (List.rev l)))
| ('S', ((h :: t) as l)) -> incr thno; thth (ios (List.hd (List.rev l)));
    List.iter thtm (List.map ios (List.tl (List.rev l)))
| ('A', [_; t]) -> incr thno; thtm (ios t)
| ('F', [_; t]) -> incr thno; thtm (ios t)
| ('Y', [_; _; _; t; s; p]) -> incr thno; thth (ios p); thtm (ios t); thtm (ios s)
| ('1', [p]) -> incr thno; thth (ios p)
| ('2', [p]) -> incr thno; thth (ios p)
| ('t', [_]) -> incr tyno
| ('a', (h :: t)) -> incr tyno; List.iter tyty (List.map ios t)
| ('v', [_; ty]) -> incr tmno; tmty (ios ty)
| ('c', [_; ty]) -> incr tmno; tmty (ios ty)
| ('f', [t; s]) -> incr tmno; tmtm (ios t); tmtm (ios s)
| ('l', [t; s]) -> incr tmno; tmtm (ios t); tmtm (ios s)
| ('+', _) -> ()
| (c, l) -> failwith ((String.make 1 c) ^ (String.concat " " l))
;;

let thth1 = Array.create size1 [];;
let thth2 = Array.create size1 [];; (* With all unused deps removed *)
let tmth = Array.create size2 [];;
let tmtm = Array.create size2 [];;
let tyth = Array.create size3 [];;
let tytm = Array.create size3 [];;
let tyty = Array.create size3 [];;
let addthth s = thth1.(s) <- !thno :: thth1.(s); thth2.(s) <- !thno :: thth2.(s);;
let addtmth s = tmth.(s) <- !thno :: tmth.(s);;
let addtyth s = tyth.(s) <- !thno :: tyth.(s);;
let addtmtm s = tmtm.(s) <- !tmno :: tmtm.(s);;
let addtytm s = tytm.(s) <- !tmno :: tytm.(s);;
let addtyty s = tyty.(s) <- !tyno :: tyty.(s);;

let process_all thth thtm tmtm thty tmty tyty =
  tyno := 0; tmno := 0; thno := 0;
  let inc = open_in "proofs" in
  try while true do
    let c = input_char inc in
    let s = input_line inc in
    process thth thtm tmtm thty tmty tyty (c, (ssplit s))
  done with End_of_file -> close_in inc;;

let count_nonnil l = Array.fold_left (fun n l -> if l = [] then n else n + 1) 0 l;;

let clean tab filter =
  for i = Array.length tab - 1 downto 1 do
    tab.(i) <- List.filter filter tab.(i)
  done;;

let clean_all () =
  clean thth2 (fun j -> not (thth2.(j) = []) || Hashtbl.mem i2f j);;
(*  clean tmth (fun j -> not (thth1.(j) = []) || Hashtbl.mem i2f j);
  clean tmtm (fun j -> not (tmth.(j) = [] && tmtm.(j) = []));
  clean tyth (fun j -> not (thth1.(j) = []) || Hashtbl.mem i2f j);
  clean tytm (fun j -> not (tmth.(j) = [] && tmtm.(j) = []));
  clean tyty (fun j -> not (tyth.(j) = [] && tytm.(j) = [] && tyty.(j) = []));;*)

read_facts ();;
process_all addthth addtmth addtmtm addtyth addtytm addtyty;;

print_string "thms: "; print_int !thno;
print_string " tms: "; print_int !tmno;
print_string " tys: "; print_int !tyno; print_newline (); flush stdout;;
print_string "Direct uses: th->th th->tm tm->tm th->ty tm->ty ty->ty: \n";
print_int (count_nonnil thth1); print_char ' ';
print_int (count_nonnil thth2); print_char ' ';
print_int (count_nonnil tmth); print_char ' ';
print_int (count_nonnil tmtm); print_char ' ';
print_int (count_nonnil tyth); print_char ' ';
print_int (count_nonnil tytm); print_char ' ';
print_int (count_nonnil tyty); print_newline (); flush stdout;;

let maxth = !thno;;
clean_all ();;

print_string "After removing:\n";
print_string "Depends: th->th th->tm tm->tm th->ty tm->ty ty->ty: \n";
print_int (count_nonnil thth1); print_char ' ';
print_int (count_nonnil thth2); print_char ' ';
print_int (count_nonnil tmth); print_char ' ';
print_int (count_nonnil tmtm); print_char ' ';
print_int (count_nonnil tyth); print_char ' ';
print_int (count_nonnil tytm); print_char ' ';
print_int (count_nonnil tyty); print_newline (); flush stdout;;

let ththr = Array.create size1 [];;
for i = 0 to size1 - 1 do
  List.iter (fun n -> if n > 0 then ththr.(n) <- i :: ththr.(n)) thth2.(i)
done;;

let rec uniq = function
  | ((x::(y::_ as t)) as l) -> let t' = uniq t in
                      if Pervasives.compare x y = 0 then t' else
                      if t'==t then l else x::t'
  | l -> l;;

let hidden_nos = Hashtbl.fold (fun i _ sf -> i :: sf) hidden [];;
let fact_nos = uniq (List.sort compare (Hashtbl.fold (fun i _ sf -> i :: sf) i2f hidden_nos));;
let rec liter2 f = function
  | h1 :: h2 :: t -> f h1 h2; liter2 f (h2 :: t)
  | _ -> ();;
let markers = Hashtbl.create 1000;;
let domark conj nextconj =
  for i = conj + 1 to nextconj - 1 do
    let mark1 fut = if fut > nextconj then Hashtbl.replace markers i () in
    List.iter mark1 thth2.(i)
  done;;
liter2 domark fact_nos;;
let brackets = List.sort compare (Hashtbl.fold (fun a b c -> a :: c) markers []);;
module IM = Map.Make(struct type t = int let compare = compare end);;
(*let max_binding m = IM.fold (fun x _ _ -> x) m 0;;
let min_binding m = IM.fold (fun x _ y -> if y <> 0 then y else x) m 0;;
*)
let max_binding m = fst (IM.max_binding m);;
let min_binding m = fst (IM.min_binding m);;

let rec rev_deps ret todo =
  if List.length ret > 1 || IM.is_empty todo then ret else
  let i = min_binding todo in
  let ntodo = IM.remove i todo in
  if (Hashtbl.mem i2f i)
  then rev_deps (i :: ret) (ntodo) else
  let ntodo = List.fold_left (fun sf e -> if e <> 0 then IM.add e () sf else sf) ntodo thth2.(i) in
  rev_deps ret ntodo;;
let l2i l = List.fold_left (fun sf e -> IM.add e () sf) IM.empty l;;
let rev_deps i = rev_deps [] (l2i thth2.(i));;
(*List.iter (fun x -> if List.length (rev_deps x) <= 1 then Hashtbl.remove markers x else print_char '.'; flush stdout) brackets;;*)
let brackets = List.sort compare (Hashtbl.fold (fun a b c -> a :: c) markers []);;
let rec deps ret todo =
  if IM.is_empty todo then IM.fold (fun k _ sf -> k :: sf) ret [] else
  let i = max_binding todo in
  let ntodo = IM.remove i todo in
  if i < Hashtbl.find f2i "ETA_AX" && i <> 8 then deps ret ntodo else
  if (Hashtbl.mem i2f i)
     (*|| Hashtbl.mem markers i*) || Hashtbl.mem hidden i
  then deps (IM.add i () ret) (ntodo) else
  let ntodo = List.fold_left (fun sf e -> IM.add e () sf) ntodo ththr.(i) in
  deps ret ntodo;;

let deps i = deps IM.empty (l2i ththr.(i));;
let bad_marker i = List.length (deps i) <= 1;;
  (*List.iter (fun i -> if bad_marker i then Hashtbl.remove markers i) brackets;;*)

  (*
let min_binding m = fst (IM.min_binding m);;

*)

print_string "Markers: ";;
flush stdout;;
let m = Hashtbl.fold (fun a b c -> a :: c) markers [];;
print_int (List.length m);;
(*print_string "->";;
List.iter (fun x -> if List.length thth2.(x) <= 1 then Hashtbl.remove markers x) m;;
let m = Hashtbl.fold (fun a b c -> a :: c) markers [];;
print_int (List.length m);;
print_string "->";;
let m = Hashtbl.fold (fun a b c -> a :: c) markers [];;
  print_int (List.length m);;*)
print_string "\n";;
flush stdout;;




let aredeps = Hashtbl.create 1000;;
List.iter (fun i -> List.iter (fun j -> Hashtbl.replace aredeps j ()) (deps i)) fact_nos;;

let outl oc tag ss is =
  let ss = ss @ (List.map string_of_int is) in
  output_char oc tag; output_string oc (String.concat " " ss); output_char oc '\n'
;;
let ty t i =
  t.(i) <- List.tl t.(i);
  if tyth.(i) = [] && tytm.(i) = [] && tyty.(i) = [] then -(i) else i;;
let tm t i =
  if t.(i) = [] && t = tmth then failwith ("tmth" ^ string_of_int i);
  t.(i) <- List.tl t.(i);
  if tmth.(i) = [] && tmtm.(i) = [] then -(i) else i;;
let th i =
  thth1.(i) <- List.tl thth1.(i);
  if thth1.(i) = [] && not (Hashtbl.mem aredeps i) then -(i) else i;;

let all_nos = ref (0 :: uniq (List.sort compare ((maxth + 1) :: Hashtbl.fold (fun i _ sf -> i :: sf) markers fact_nos)));;

let out_plus oc thno prevno =
  let n = Hashtbl.find i2f thno in
  let ds = deps thno in
  if Hashtbl.mem hidden thno then outl oc '+' ["ignore_axiom"; n] [] else
  if ds = [] then outl oc '+' ["ignore_trivial"; n] [] else
  if List.exists (fun i -> Hashtbl.mem hidden i && not (Hashtbl.mem i2f i)) ds then
    outl oc '+' ["ignore_definition"; n] [] else
  let ns = List.map (fun x -> string_of_int x ^ " " ^ Hashtbl.find i2f x) ds  in
    outl oc '+' ((if Hashtbl.mem aredeps thno then "train" else "test") :: n :: string_of_int (thno - prevno) :: string_of_int (List.length ds) :: ns) []
;;

let outt oc tag ss tys tms ths =
  let prevno = List.hd !all_nos in
  if !thno = 1 + prevno then begin
    all_nos := List.tl !all_nos;
    let curno = List.hd !all_nos in
    try out_plus oc curno prevno with Not_found ->
    if curno = maxth + 1 || Hashtbl.mem hidden curno then outl oc '+' ["-"] [] (* Ignore *)
    else if Hashtbl.mem markers curno then outl oc '+' ["-"] [] else failwith ("sth: " ^ string_of_int !thno)
  end;
  if thth2.(!thno) = [] && not (Hashtbl.mem i2f !thno) then output_string oc "-";
  outl oc tag ss
    (List.map (fun i -> ty tyth (ios i)) tys @ List.map
      (fun i -> tm tmth (ios i)) tms @ List.map (fun i -> th (ios i)) ths);
;;

(*    if Hashtbl.mem tags !thno then outl oc '+' ["!"] []*)

let outtm oc tag ss tys tms =
  (outl oc tag ss (List.map (fun i -> ty tytm (ios i)) tys @ List.map (fun i -> tm tmtm (ios i)) tms))
;;

let outty oc tag ss tys =
  (outl oc tag ss (List.map (fun i -> ty tyty (ios i)) tys))
;;

let printer oc = function
  ('R', [t]) -> incr thno; outt oc 'R' [] [] [t] []
| ('B', [t]) -> incr thno; outt oc 'B' [] [] [t] []
| ('T', [p; q]) -> incr thno; outt oc 'T' [] [] [] [p; q]
| ('C', [p; q]) -> incr thno; outt oc 'C' [] [] [] [p; q]
| ('L', [t; p]) -> incr thno; outt oc 'L' [] [] [t] [p]
| ('H', [t]) -> incr thno; outt oc 'H' [] [] [t] []
| ('E', [p; q]) -> incr thno; outt oc 'E' [] [] [] [p; q]
| ('D', [p; q]) -> incr thno; outt oc 'D' [] [] [] [p; q]
| ('Q', ((h :: t) as l)) -> incr thno;
    let (th, tys) = (List.hd (List.rev l), List.rev (List.tl (List.rev l))) in
    outt oc 'Q' [] tys [] [th]
| ('S', ((h :: t) as l)) -> incr thno;
    let (th, tms) = (List.hd (List.rev l), List.rev (List.tl (List.rev l))) in
    outt oc 'S' [] [] tms [th]
| ('A', [n; t]) -> incr thno; outt oc 'A' [n] [] [t] []
| ('F', [n; t]) -> incr thno; outt oc 'F' [n] [] [t] []
| ('Y', [n1; n2; n3; t; s; p]) -> incr thno; outt oc 'Y' [n1; n2; n3] [] [t; s] [p]
| ('1', [p]) -> incr thno; outt oc '1' [] [] [] [p]
| ('2', [p]) -> incr thno; outt oc '2' [] [] [] [p]
| ('t', [n]) -> incr tyno; outty oc 't' [n] []
| ('a', (h :: t)) -> incr tyno; outty oc 'a' [h] t
| ('v', [n; ty]) -> incr tmno; outtm oc 'v' [n] [ty] []
| ('c', [n; ty]) -> incr tmno; outtm oc 'c' [n] [ty] []
| ('f', [t; s]) -> incr tmno; outtm oc 'f' [] [] [t; s]
| ('l', [t; s]) -> incr tmno; outtm oc 'l' [] [] [t; s]
| ('+', _) -> ()
| (c, l) -> failwith ((String.make 1 c) ^ (String.concat " " l))
;;

let print_all () =
  tyno := 0; tmno := 0; thno := 0;
  let inc = open_in "proofs" in
  let oc = open_out "proofsN" in
  try while true do
      let c = input_char inc in
      let s = input_line inc in
      printer oc (c, ssplit s);
      flush oc
    done
  with End_of_file -> (close_in inc; outl oc '+' ["="] []; close_out oc);;

print_all ();;
